#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#

from distutils.core import setup

setup(
        name = 'on_the_fly',
        packages = ['on_the_fly',],
        version = "0.01rc",
        description = 'easy machine learning',
        author = 'Peng Yu',
        author_email = 'yupbank@gmail.com',
        url = 'https://github.com/yupbank/on_the_fly',
        download_url = 'https://github.com/yupbank/on_the_fly/archive/0.01.tar.gz',
        keywords = ['on the fly', 'machine learning', 'sklearn', 'online'],
        classifiers = [],
        )
